searchData={"items":[{"type":"module","title":"Chorex","doc":"Make your modules dance!\n\nChorex allows you to specify a choreography: a birds-eye view of an\ninteraction of concurrent parties. Chorex takes that choreography\ncreates a *projection* of that interaction for each party in the\nsystem.\n\nTake, for example, the classic problem of a book seller and two\nbuyers who want to split the price. The interaction looks like this:\n\n```\n+------+         +------+ +------+\n|Buyer1|         |Seller| |Buyer2|\n+--+---+         +--+---+ +--+---+\n   |                |        |\n   |   Book title   |        |\n   |--------------->|        |\n   |                |        |\n   |     Price      |        |\n   |<---------------|        |\n   |                |        |\n   |                |  Price |\n   |                |------->|\n   |                |        |\n   |      Contribution       |\n   |<------------------------|\n   |                |        |\n   |   Buy/No buy   |        |\n   |--------------->|        |\n   |                |        |\n   |(if Buy) address|        |\n   |--------------->|        |\n   |                |        |\n   | Shipping date  |        |\n   |<---------------|        |\n+--+---+         +--+---+ +--+---+\n|Buyer1|         |Seller| |Buyer2|\n+------+         +------+ +------+\n```\n\nYou can encode that interaction with the `defchor` macro and DSL:\n\n```elixir\ndefmodule ThreePartySeller do\n  defchor [Buyer1, Buyer2, Seller] do\n    Buyer1.get_book_title() ~> Seller.(b)\n    Seller.get_price(\"book:\" <> b) ~> Buyer1.(p)\n    Seller.get_price(\"book:\" <> b) ~> Buyer2.(p)\n    Buyer2.compute_contrib(p) ~> Buyer1.(contrib)\n\n    if Buyer1.(p - contrib   Seller\n      Buyer1.get_address() ~> Seller.(addr)\n      Seller.get_delivery_date(b, addr) ~> Buyer1.(d_date)\n      Buyer1.(d_date)\n    else\n      Buyer1[R] ~> Seller\n      Buyer1.(nil)\n    end\n  end\nend\n```\n\nThe `defchor` macro will take care of generating code that handles\nsending messages. Now all you have to do is implement the local\nfunctions that don't worry about the outside system:\n\n```elixir\ndefmodule Seller do\n  use ThreePartySeller.Chorex, :seller\n\n  def get_price(book_name), do: ...\n  def get_delivery_date(book_name, addr), do: ...\nend\n\ndefmodule Buyer1 do\n  use ThreePartySeller.Chorex, :buyer1\n\n  def get_book_title(), do: ...\n  def get_address(), do: ...\n  def get_budget(), do: ...\nend\n\ndefmodule Buyer2 do\n  use ThreePartySeller.Chorex, :buyer2\n\n  def compute_contrib(price), do: ...\nend\n```\n\nWhat the `defchor` macro actually does is creates a module `Chorex`\nand submodules for each of the actors: `Chorex.Buyer1`,\n`Chorex.Buyer2` and `Chorex.Seller`. There's a handy `__using__`\nmacro that will Do the Right Thing™ when you say `use Mod.Chorex, :actor_name`\nand will import those modules and say that your module implements\nthe associated behaviour. That way, you should get a nice\ncompile-time warning if a function is missing.\n\nTo start the choreography, you need to invoke the `init` function in\neach of your actors (provided via the `use ...` invocation)\nwhereupon each actor will wait to receive a config mapping actor\nname to PID:\n\n```elixir\nthe_seller = spawn(MySeller, :init, [])\nthe_buyer1 = spawn(MyBuyer1, :init, [])\nthe_buyer2 = spawn(MyBuyer2, :init, [])\n\nconfig = %{Seller1 => the_seller, Buyer1 => the_buyer1, Buyer2 => the_buyer2, :super => self()}\n\nsend(the_seller, {:config, config})\nsend(the_buyer1, {:config, config})\nsend(the_buyer2, {:config, config})\n\nassert_receive {:chorex_return, Buyer1, ~D[2024-05-13]}\n```\n\nEach of the parties will try sending the last value they computed\nonce they're done running.","ref":"Chorex.html"},{"type":"module","title":"Higher-order choreographies - Chorex","doc":"Chorex supports higher-order choreographies. For example, you can\ndefine a generic buyer/seller interaction and abstract away the\ndecision process into a higher-order choreography:\n\n```elixir\ndefmodule TestChor3 do\n  defchor [Buyer3, Contributor3, Seller3] do\n    def bookseller(decision_func) do\n      Buyer3.get_book_title() ~> Seller3.the_book\n      with Buyer3.decision <- decision_func.(Seller3.get_price(\"book:\" <> the_book)) do\n        if Buyer3.decision do\n          Buyer3[L] ~> Seller3\n          Buyer3.get_address() ~> Seller3.the_address\n          Seller3.get_delivery_date(the_book, the_address) ~> Buyer3.d_date\n          Buyer3.d_date\n        else\n          Buyer3[R] ~> Seller3\n          Buyer3.(nil)\n        end\n      end\n    end\n\n    def one_party(Seller3.(the_price)) do\n      Seller3.(the_price) ~> Buyer3.(p)\n      Buyer3.(p   Buyer3.(p)\n      Seller3.(the_price) ~> Contributor3.(p)\n      Contributor3.compute_contrib(p) ~> Buyer3.(contrib)\n      Buyer3.(p - contrib < get_budget())\n    end\n\n    bookseller(&two_party/1)\n  end\nend\n```\n\nThis will run the two-buyer scenario by default. If you want to cut\nthe second buyer out of the picture, define a function called\n`run_choreography` for the buyer and seller actors and have them\ncompose the `one_party` and `bookseller` functions.\n\n```elixir\ndefmodule MySeller31 do\n  use TestChor3.Chorex, :seller3\n\n  def get_delivery_date(_book, _addr) do\n    ~D[2024-05-13]\n  end\n\n  def get_price(\"book:Das Glasperlenspiel\"), do: 42\n  def get_price(\"book:Zen and the Art of Motorcycle Maintenance\"), do: 13\n\n  def run_choreography(impl, config) do\n    Seller3.bookseller(impl, config, &Seller3.one_party/3)\n  end\nend\n\ndefmodule MyBuyer31 do\n  use TestChor3.Chorex, :buyer3\n\n  def get_book_title(), do: \"Zen and the Art of Motorcycle Maintenance\"\n  def get_address(), do: \"Maple Street\"\n  def get_budget(), do: 22\n\n  def run_choreography(impl, config) do\n    Buyer3.bookseller(impl, config, &Buyer3.one_party/3)\n  end\nend\n```\n\nIt's important to remember to pass `impl` and `config` around. These\nare internal to the workings of the Chorex module, so do not modify them.","ref":"Chorex.html#module-higher-order-choreographies"},{"type":"module","title":"Singletons managing shared state - Chorex","doc":"Sometimes, you might want to share some state between different\ninstances of the same choreography. The classic Elixir solution to\nmanaging shared state is to use a GenServer: processes interested in\naccessing/modifying the state send messages to the GenServer and await\nreplies.\n\nChorex provides a mechanism to model this behavior in a choreography.\nGoing back to our bookseller example, suppose there is a limited stock\nof books, and the seller must not sell a book twice. The stock of\nbooks is the shared state, and instances of the seller in the\nchoreography need to be able to access this.\n\nHere is how you define such a choreography:\n\n```elixir\ndefchor [Buyer, {Seller, :singleton}] do\n  Buyer.get_book_title() ~> Seller.(b)\n  Seller.get_price(b) ~> Buyer.(p)\n  if Buyer.in_budget(p) do\n    Buyer[L] ~> Seller\n    if Seller.acquire_book(@chorex_config, b) do\n      Seller[L] ~> Buyer\n      Buyer.(:book_get)\n    else\n      Seller[R] ~> Buyer\n      Buyer.(:darn_missed_it)\n    end\n  else\n    Buyer[R] ~> Seller\n    Buyer.(:nevermind)\n  end\nend\n```\n\nSaying `{Seller, :singleton}` in the `defchor` declaration indicates\nthat the `Seller` actor is going to share some state. The `Seller`\nactor can access this shared state in any function, though such\nfunctions need to have the magic `@chorex_config` variable passed to\nthem. (This is just a special symbol recognized by the Chorex\ncompiler.)\n\nIn the implementation, the Seller can access the state using the\n`Proxy.update_state` function:\n\n```elixir\ndefmodule MySellerBackend do\n  use BooksellerProxied.Chorex, :seller\n  alias Chorex.Proxy\n\n  def get_price(_), do: 42\n\n  def acquire_book(config, book_title) do\n\n    # Attempt to acquire a lock on the book\n    Proxy.update_state(config, fn book_stock ->\n      with {:ok, count} <- Map.fetch(book_stock, book_title) do\n        if count > 0 do\n          # Have the book, lock it for this customer\n          {true, Map.put(book_stock, book_title, count - 1)}\n        else\n          {false, book_stock}\n        end\n      else\n        :error ->\n          {false, book_stock}\n      end\n    end)\n  end\nend\n```\n\nThat's it! Now the seller won't accidentally double-sell a book.","ref":"Chorex.html#module-singletons-managing-shared-state"},{"type":"module","title":"The need for a proxy - Chorex","doc":"Actors that share state do run as a separate process, but a GenServer\nthat manages the state also acts as a proxy for all messages to/from\nthe actor. This is so that operations touching the shared state happen\nin lockstep with progression through the choreography. We may\ninvestigate weakening this property in the future.","ref":"Chorex.html#module-the-need-for-a-proxy"},{"type":"module","title":"Setting up the shared-state choreography - Chorex","doc":"You need to be a little careful when setting up the shared state\nchoreography. Instead of setting up all the actors manually, you need\nto set up *one* instance of each shared-state actor, then create\nseparate *sessions* for each instance of the choreography that you\nwant to run.\n\nHere is an example with two buyers trying to buy the same book:\n\n```elixir\n# Start up the buyers\nb1 = spawn(MyBuyer, :init, [])\nb2 = spawn(MyBuyer, :init, [])\n\n# Start up the seller proxy with the initial shared\n# state (the stock of books in this case)\n{:ok, px} = GenServer.start(Chorex.Proxy, %{\"Anathem\" => 1})\n\n# Start sessions: one for each buyer\nProxy.begin_session(px, [b1], MySellerBackend, :init, [])\nconfig1 = %{Buyer => b1, Seller => px, :super => self()}\n\nProxy.begin_session(px, [b2], MySellerBackend, :init, [])\nconfig2 = %{Buyer => b2, Seller => px, :super => self()}\n\n# Send everyone their configuration\nsend(b1, {:config, config1})\nsend(px, {:chorex, b1, {:config, config1}})\nsend(b2, {:config, config2})\nsend(px, {:chorex, b2, {:config, config2}})\n```\n\nThe `Proxy.begin_sesion` function takes a proxy function, a list of\nPIDs that partake in a given session, and a module, function, arglist\nfor the thing to proxy.\n\n**Sessions**: PIDs belonging to a session will have their messages\nrouted to the corresponding proxied process. The GenServer looks up\nwhich session a PID belongs to, finds the proxied process linked to\nthat session, then forwards the message to that process. The exact\nmechanisms of how this works may change in the future to accommodate\nrestarts.\n\nWhen you send the config information to a proxied process, you send it\nthrough the proxy first, and you must wrap the message as shown above\nwith a process from the session you want to target as the second\nelement in the tuple; this just helps the proxy figure out the session\nyou want.\n\nThat's it! If you run the above choreography, the process that kicks\nthis all off will get *one* message like `{:chorex_return, Buyer, :book_get}`\nand *one* message like `{:chorex_return, Buyer, :darn_missed_it}`,\nindicating that exactly one of the buyers got the coveted book.","ref":"Chorex.html#module-setting-up-the-shared-state-choreography"},{"type":"function","title":"Chorex.actor_from_local_exp/2","doc":"Get the actor name from an expression\n\n    iex> Chorex.actor_from_local_exp((quote do: Foo.bar(42)), __ENV__)\n    {:ok, Foo}","ref":"Chorex.html#actor_from_local_exp/2"},{"type":"macro","title":"Chorex.defchor/2","doc":"Define a new choreography.","ref":"Chorex.html#defchor/2"},{"type":"function","title":"Chorex.do_local_project_wrapper/5","doc":"","ref":"Chorex.html#do_local_project_wrapper/5"},{"type":"function","title":"Chorex.empty_ctx/0","doc":"","ref":"Chorex.html#empty_ctx/0"},{"type":"function","title":"Chorex.flatten_block/1","doc":"","ref":"Chorex.html#flatten_block/1"},{"type":"macro","title":"Chorex.is_immediate/1","doc":"","ref":"Chorex.html#is_immediate/1"},{"type":"function","title":"Chorex.merge/2","doc":"Perform the control merge function, but flatten block expressions at each step","ref":"Chorex.html#merge/2"},{"type":"function","title":"Chorex.merge_step/2","doc":"","ref":"Chorex.html#merge_step/2"},{"type":"function","title":"Chorex.project/4","doc":"Perform endpoint projection in the context of node `label`.\n\nThis returns a pair of a projection for the label, and a list of\nbehaviors that an implementer of the label must implement.\n\nArguments:\n\n1. Elixir AST term to project.\n2. Macro environment.\n3. Name of the actor currently under projection. Atom.\n4. Extra information about the expansion. Map. Currently contains\n   just a list of actors that will be behind a proxy.\n\nReturns an instance of the `WriterMonad`, which is just a 3-tuple\ncontaining:\n\n1. The projected term. Elixir AST.\n2. A list of callback specifications for this actor. (Functions the\n   actor implementer needs to have.)\n3. List of auxiliary functions generated during the projection process.","ref":"Chorex.html#project/4"},{"type":"function","title":"Chorex.project_global_func/5","doc":"","ref":"Chorex.html#project_global_func/5"},{"type":"function","title":"Chorex.project_local_expr/4","doc":"Like `project/3`, but focus on handling `ActorName.local_var`,\n`ActorName.local_func()` or `ActorName.(local_exp)`. Handles walking\nthe local expression to gather list of functions needed for the\nbehaviour to implement.","ref":"Chorex.html#project_local_expr/4"},{"type":"function","title":"Chorex.project_local_func/5","doc":"","ref":"Chorex.html#project_local_func/5"},{"type":"function","title":"Chorex.project_sequence/4","doc":"","ref":"Chorex.html#project_sequence/4"},{"type":"function","title":"Chorex.walk_local_expr/4","doc":"","ref":"Chorex.html#walk_local_expr/4"},{"type":"module","title":"Chorex.Proxy","doc":"Proxy for singletons in a choreography.","ref":"Chorex.Proxy.html"},{"type":"function","title":"Chorex.Proxy.begin_session/5","doc":"","ref":"Chorex.Proxy.html#begin_session/5"},{"type":"function","title":"Chorex.Proxy.child_spec/1","doc":"Returns a specification to start this module under a supervisor.\n\nSee `Supervisor`.","ref":"Chorex.Proxy.html#child_spec/1"},{"type":"function","title":"Chorex.Proxy.fetch_state/1","doc":"","ref":"Chorex.Proxy.html#fetch_state/1"},{"type":"function","title":"Chorex.Proxy.init/1","doc":"","ref":"Chorex.Proxy.html#init/1"},{"type":"function","title":"Chorex.Proxy.send_proxied/2","doc":"Send a message to a proxied service.\n\nHandles the wrapping of the message with the `{:chorex, self(), ...}`\ntuple so that the proxy knows which session to send the message on to.","ref":"Chorex.Proxy.html#send_proxied/2"},{"type":"function","title":"Chorex.Proxy.set_state/2","doc":"","ref":"Chorex.Proxy.html#set_state/2"},{"type":"function","title":"Chorex.Proxy.update_state/2","doc":"Updates the session state for the current process.\n\nFor use by proxied processes.","ref":"Chorex.Proxy.html#update_state/2"},{"type":"type","title":"Chorex.Proxy.session_key/0","doc":"","ref":"Chorex.Proxy.html#t:session_key/0"},{"type":"type","title":"Chorex.Proxy.session_state/0","doc":"","ref":"Chorex.Proxy.html#t:session_state/0"},{"type":"type","title":"Chorex.Proxy.state/0","doc":"","ref":"Chorex.Proxy.html#t:state/0"},{"type":"behaviour","title":"Customer","doc":"","ref":"Customer.html"},{"type":"callback","title":"Customer.allergic_to/2","doc":"","ref":"Customer.html#c:allergic_to/2"},{"type":"function","title":"Customer.big_chor/3","doc":"","ref":"Customer.html#big_chor/3"},{"type":"function","title":"Customer.chorex_send/2","doc":"","ref":"Customer.html#chorex_send/2"},{"type":"callback","title":"Customer.get_allergens/0","doc":"","ref":"Customer.html#c:get_allergens/0"},{"type":"callback","title":"Customer.get_bread/0","doc":"","ref":"Customer.html#c:get_bread/0"},{"type":"function","title":"Customer.hamncheese/3","doc":"","ref":"Customer.html#hamncheese/3"},{"type":"function","title":"Customer.init/1","doc":"","ref":"Customer.html#init/1"},{"type":"function","title":"Customer.pbj/3","doc":"","ref":"Customer.html#pbj/3"},{"type":"callback","title":"Customer.plz_wash/0","doc":"","ref":"Customer.html#c:plz_wash/0"},{"type":"function","title":"Customer.run_choreography/2","doc":"","ref":"Customer.html#run_choreography/2"},{"type":"module","title":"CustomerAlice","doc":"","ref":"CustomerAlice.html"},{"type":"function","title":"CustomerAlice.allergic_to/2","doc":"","ref":"CustomerAlice.html#allergic_to/2"},{"type":"function","title":"CustomerAlice.get_allergens/0","doc":"","ref":"CustomerAlice.html#get_allergens/0"},{"type":"function","title":"CustomerAlice.get_bread/0","doc":"","ref":"CustomerAlice.html#get_bread/0"},{"type":"function","title":"CustomerAlice.init/0","doc":"","ref":"CustomerAlice.html#init/0"},{"type":"function","title":"CustomerAlice.plz_wash/0","doc":"","ref":"CustomerAlice.html#plz_wash/0"},{"type":"function","title":"CustomerAlice.run_choreography/2","doc":"","ref":"CustomerAlice.html#run_choreography/2"},{"type":"module","title":"MyStoreBackendImpl","doc":"","ref":"MyStoreBackendImpl.html"},{"type":"function","title":"MyStoreBackendImpl.make_sandwich/2","doc":"","ref":"MyStoreBackendImpl.html#make_sandwich/2"},{"type":"behaviour","title":"StoreBackend","doc":"","ref":"StoreBackend.html"},{"type":"function","title":"StoreBackend.big_chor/3","doc":"","ref":"StoreBackend.html#big_chor/3"},{"type":"function","title":"StoreBackend.hamncheese/3","doc":"","ref":"StoreBackend.html#hamncheese/3"},{"type":"function","title":"StoreBackend.init/1","doc":"","ref":"StoreBackend.html#init/1"},{"type":"callback","title":"StoreBackend.make_sandwich/2","doc":"","ref":"StoreBackend.html#c:make_sandwich/2"},{"type":"function","title":"StoreBackend.pbj/3","doc":"","ref":"StoreBackend.html#pbj/3"},{"type":"function","title":"StoreBackend.run_choreography/2","doc":"","ref":"StoreBackend.html#run_choreography/2"},{"type":"module","title":"StoreProxy","doc":"","ref":"StoreProxy.html"},{"type":"function","title":"StoreProxy.child_spec/1","doc":"Returns a specification to start this module under a supervisor.\n\nSee `Supervisor`.","ref":"StoreProxy.html#child_spec/1"},{"type":"function","title":"StoreProxy.init/1","doc":"","ref":"StoreProxy.html#init/1"},{"type":"type","title":"StoreProxy.session_key/0","doc":"","ref":"StoreProxy.html#t:session_key/0"},{"type":"type","title":"StoreProxy.state/0","doc":"","ref":"StoreProxy.html#t:state/0"},{"type":"module","title":"Utils","doc":"","ref":"Utils.html"},{"type":"function","title":"Utils.downcase_atom/1","doc":"iex> Utils.downcase_atom(Foo)\n:foo","ref":"Utils.html#downcase_atom/1"},{"type":"function","title":"Utils.fresh_atom/1","doc":"","ref":"Utils.html#fresh_atom/1"},{"type":"function","title":"Utils.upcase_atom/1","doc":"iex> Utils.upcase_atom(:foo)\n:Foo","ref":"Utils.html#upcase_atom/1"},{"type":"module","title":"WriterMonad","doc":"Makes it easy to do operations on code and track gathered data.","ref":"WriterMonad.html"},{"type":"function","title":"WriterMonad.<~>/2","doc":"","ref":"WriterMonad.html#%3C~%3E/2"},{"type":"function","title":"WriterMonad.bind/2","doc":"","ref":"WriterMonad.html#bind/2"},{"type":"function","title":"WriterMonad.fmap/2","doc":"","ref":"WriterMonad.html#fmap/2"},{"type":"function","title":"WriterMonad.mapM/2","doc":"","ref":"WriterMonad.html#mapM/2"},{"type":"macro","title":"WriterMonad.monadic/1","doc":"Haskell-like `do` notation for monads.\n\niex> import WriterMonad\niex> monadic do\n...>   thing1 <- return(5)\n...>   thing2 <- {thing1 + 2, [\"foo\"], [:zoop]}\n...>   thing3 <- {thing1 + thing2, [\"bar\"], [:quux]}\n...>   return thing3\n...> end\n{12, [\"bar\", \"foo\"], [:zoop, :quux]}","ref":"WriterMonad.html#monadic/1"},{"type":"function","title":"WriterMonad.mzero/0","doc":"","ref":"WriterMonad.html#mzero/0"},{"type":"function","title":"WriterMonad.return/1","doc":"","ref":"WriterMonad.html#return/1"},{"type":"function","title":"WriterMonad.return/2","doc":"","ref":"WriterMonad.html#return/2"},{"type":"function","title":"WriterMonad.return/3","doc":"","ref":"WriterMonad.html#return/3"},{"type":"function","title":"WriterMonad.~>>/2","doc":"","ref":"WriterMonad.html#~%3E%3E/2"},{"type":"type","title":"WriterMonad.t/0","doc":"","ref":"WriterMonad.html#t:t/0"},{"type":"exception","title":"Chorex.ProjectionError","doc":"","ref":"Chorex.ProjectionError.html"},{"type":"extras","title":"README","doc":"Chorex - Choreographic Programming in Elixir\n\n\n# Synopsis\n\n**Note:** this documentation is current as of 2024-05-30. The project is evolving rapidly, so this README may occasionally get out-of-sync with what the project can do.\n\nDescribe the choreography in a module with the `defchor` macro:\n\n```elixir\ndefmodule TestChor do\n  defchor [Buyer, Seller] do\n    Buyer.get_book_title() ~> Seller.(b)\n    Seller.get_price(b) ~> Buyer.(p)\n    Buyer.(p)\n  end\nend\n\n```\n\nImplement the actors:\n\n```elixir\ndefmodule MyBuyer do\n  use TestChor.Chorex, :buyer\n\n  def get_book_title(), do: \"Das Glasperlenspiel\"\nend\n\ndefmodule MySeller do\n  use TestChor.Chorex, :seller\n\n  def get_price(_b), do: 42\nend\n```\n\nElsewhere in your program:\n\n```elixir\nps = spawn(MySeller, :init, [])\npb = spawn(MyBuyer, :init, [])\n\nconfig = %{Seller => ps, Buyer => pb, :super => self()}\n\nsend(ps, {:config, config})\nsend(pb, {:config, config})\n\nreceive do\n  {:chorex_return, Buyer, val} -> IO.puts(\"Got #{val}\")\nend\n```\n\nThe program should print `Got 42` to the terminal.\n\n\n# Description\n\nChorex is a library for *choreographic programming* in Elixir. Choreographic programming is a programming paradigm where you specify the interactions between different entities in a concurrent system in one global view, and then *extract implementations* for each of those actors. See [§ Bibliography](#org44e5aee) for references on choreographic programming in general.","ref":"readme.html"},{"type":"extras","title":"Installation - README","doc":"Chorex is available on Hex.pm. Install by including the following in your `mix.exs` file under the `deps` list:\n\n```elixir\ndef deps do\n  [\n    ...,\n    {:chorex, \"~> 0.1.0\"},\n    ...\n  ]\nend\n```\n\nYou can install development versions of Chorex directly from GitHub like so:\n\n```elixir\ndef deps do\n  [\n    ...,\n    {:chorex, github: \"utahplt/chorex\"},\n    ...\n  ]\nend\n```\n\nNote that this is *experimental software* and stuff *will* break. Please don't rely on this for anything production-grade. Not yet at least.","ref":"readme.html#installation"},{"type":"extras","title":"What is a choreography? - README","doc":"A choreography is a birds-eye view of an interaction between nodes in a distributed system. You have some set of *actors/—in Elixir parlance processes—that exchange /messages* while also running some /local computation/—i.e. functions that don't rely on talking to other nodes in the system.","ref":"readme.html#what-is-a-choreography"},{"type":"extras","title":"Choreography syntax - README","doc":"Chorex introduces some new syntax for choreographies. Here's a breakdown of how it works:\n\n```elixir\ndefchor [Actor1, Actor2, ...] do\n  ...choreography body...\nend\n```\n\nThe `defchor` macro wraps a choreography and translates it into core Elixir code. You give `defchor` a list of actors, specified as if they were module names, and then a `do` block wraps the choreography body.\n\n```elixir\nActor1.(var1) ~> Actor2.(var2_a)\nActor1.func_1() ~> Actor2.(var2_b)\nActor1.func_2(var1_a, var1_b) ~> Actor2.(var2_c)\nActor1.(var1_a + var1_b) ~> Actor2.(var2_c)\n```\n\nThe `~>` indicates sending a message between actors. The left-hand-side must be `Actor1. `, where that ` ` bit can be one of three things:\n\n1.  A variable local to Actor1\n2.  A function local to Actor1 (with or without arguments, also all local to Actor1)\n3.  An expression local to Actor1\n\nThe right-and-side must be `Actor2. `. This means that the left-hand-side will be computed on `Actor1` and send to `Actor2` where it will be stored in variable ` `.\n\n**ACHTUNG!! `mix format` will rewrite `Actor1.var1` to `Actor1.var1()` which is a function call instead of a variable! Wrap variables in parens like `Actor1.(var1)` if you want to use `mix format`!** This is an unfortunate drawback—suggestions on fixing this would be welcome.\n\nLocal functions are not defined as part of the choreography; instead, you implement these in a separate Elixir module. More on that later.\n\n```elixir\nif Actor1.make_decision() do\n  Actor1[L] ~> Actor2\n  ...\nelse\n  Actor1[R] ~> Actor2\n  ...\nend\n```\n\n`if` expressions are supported. Some actor makes a choice of which branch to go down. It is then *crucial* (and, at this point, entirely up to the user) that that deciding actor inform all other actors about the choice of branch with the special `ActorName[L] ~> OtherActorName` syntax. Note the lack of `.` and variable names. Furthermore, the true branch is always `L` (left) and the false branch is always `R` (right).\n\n```elixir\ndef higher_order_chor(other_chor) do\n  ... other_chor.(...) ...\nend\n```\n\nChorex supports higher-order choreographies. These are choreographies that take another choreography as an argument where it can be applied like a function.\n\n```elixir\ndef some_local_chor(Actor.(var_name)) do\n  Actor.(var_name) ~> OtherActor.(other_var)\n  OtherActor.(other_var)\nend\n```\n\nThis creates a choreography that can be passed as an argument to the `higher_order_chor` function. This takes as an argument a variable living at a particular actor, and returns another value on a potentially different node.\n\nYou would combine the choreographies like so:\n\n```elixir\ndefchor [Actor, OtherActor] do\n  def higher_order_chor(other_chor) do\n    ... other_chor.(...) ...\n  end\n\n  def some_local_chor(Actor.(var_name)) do\n    Actor.(var_name) ~> OtherActor.(other_var)\n    OtherActor.(other_var)\n  end\n\n  higher_order_chor(&some_local_chor/1)\nend\n```\n\nRight now these functions are limited to a single argument.\n\n```elixir\nwith OtherActor.(other_var) <- other_chor.(Actor.(var)) do\n  ...\nend\n```\n\nYou can use `with` to bind a variable to the result of calling a higher-order choreography. Note that right now you can only have one `<-` in the expression.","ref":"readme.html#choreography-syntax"},{"type":"extras","title":"Creating a choreography - README","doc":"To create a choreography, start by making a module, and writing the choreography with the `defchor` macro.\n\n```elixir\ndefmodule Bookstore do\n  defchor [Actor1, Actor2] do\n    Actor1.(... some expr ...) ~> Actor2.(some_var)\n    Actor2.some_computation(some_var) ~> Actor1.(the_result)\n    ...\n  end\nend\n```\n\nYou will need to make a module for every actor you specify at the beginning of `defchor` and mark which actor you're implementing like so:\n\n```elixir\ndefmodule MyFirstActor do\n  use Bookstore.Chorex, :actor1\n\n  ...\nend\n\ndefmodule MySecondActor do\n  use Bookstore.Chorex, :actor2\n\n  def some_computation(val), do: ...\nend\n```\n\nThese modules will need to implement all of the local functions specified in the choreography. Chorex will use Elixir's behaviour mechanism to warn you if you don't implement every function needed. In the above example, the `MySecondActor` implements the role of `Actor2` in the choreography, and therefore needs to implement the `some_computation` function.\n\n**Note:** *Actor names do not need to be the same as the modules implementing them!* It is *useful* to do that, but there exist instances where you might want to write one choreography and implement it in different ways.","ref":"readme.html#creating-a-choreography"},{"type":"extras","title":"Running a choreography - README","doc":"To fire off the choreography, you need to spin up a process for each actor and then tell each actor where to find the other actors in the system. For the above example, you could do this:\n\n```elixir\nfirst_actor = spawn(MyFirstActor, :init, [])\nsecond_actor = spawn(MySecondActor, :init, [])\n\nconfig = %{Actor1 => first_actor, Actor2 => second_actor, :super => self()}\nsend(first_actor, config)\nsend(second_actor, config)\n```\n\nOnce the actors are done, they will send the last value they computed to `:super` tagged with the actor they were implementing. So, for this example, you could see what `Actor1` computed by awaiting:\n\n```elixir\nreceive do\n  {:chorex_return, Actor1, val} -> IO.inspect(val, label: \"Actor1's return: \")\nend\n```","ref":"readme.html#running-a-choreography"},{"type":"extras","title":"Shared-state choreographies - README","doc":"Sometimes you might have a choreography where one or more actors need to share some state between different instantiations of the choreography. Returning to our bookseller example, the bookseller might need to keep track of a finite stock of books and ensure that no book gets double-sold.\n\nChorex can let you share state between different instances of the bookseller actor through a proxy. Details are under the `Chorex` module.","ref":"readme.html#shared-state-choreographies"},{"type":"extras","title":"Using a choreography with the rest of your project - README","doc":"The local functions are free to call any other code you have—they're just normal Elixir. If that code sends and receives messages not managed by the choreography library, there is no guarantee that this will be deadlock-free.\n\n\n# Development\n\nChorex is under active development and things will change and break rapidly.\n\nIf you find any bugs or would like to suggest a feature, please [open an issue on GitHub](https://github.com/utahplt/chorex/issues).","ref":"readme.html#using-a-choreography-with-the-rest-of-your-project"},{"type":"extras","title":"Changelog - README","doc":"We will collect change descriptions here until we come up with a more stable format when changes get bigger.\n\n - v0.2.0; 2024-07-03\n \n   Add shared-state actors.\n\n - v0.1.0; 2024-05-30\n    \n   Initial release. Lots of rough edges so please, be patient. :)","ref":"readme.html#changelog"},{"type":"extras","title":"High-level internals - README","doc":"The `defchor` macro is implemented in the `Chorex` module.\n\n-   The `defchor` macro gathers a list of actors.\n-   For each actor, call `project` on the body of the choreography. The `project` function keeps track of the current actor as the `label` variable. (This vernacular borrowed from the academic literature.)\n-   The functions `project` and `project_sequence` are mutually recursive: `project_sequence` gets invoked whenever `project` encounters a block with multiple instructions.\n-   The `project` function walks the AST, it gathers a list of functions that will need to be implemented by each actor's implementing module, as well as a list of top-level functions for each projection.\n    -   This gathering is handled by the `WriterMonad` module, which provides the `monadic do ... end` form as well as `return` and `mzero`.\n-   Finally the macro generates modules for each actor under the `Chorex` module it generates.\n\nSo, for example, if you have a simple Choreography like this:\n\n```elixir\ndefchor [Alice, Bob] do\n  Alice.pick_modulus() ~> Bob.(m)\n  Bob.gen_key(m) ~> Alice.(bob_key)\n  Alice.encrypt(message, bob_key)\nend\n```\n\nThis will get transformed into (roughly) this code:\n\n```elixir\ndefmodule Chorex do\n  (\n    def alice do\n      quote do\n        import Alice\n        @behaviour Alice\n        def init() do\n          Alice.init(__MODULE__)\n        end\n      end\n    end\n\n    defmodule Alice do\n      @callback encrypt(any(), any()) :: any()\n      @callback pick_modulus() :: any()\n      def init(impl) do\n        receive do\n          {:config, config} ->\n            ret = run_choreography(impl, config)\n            send(config[:super], {:chorex_return, Alice, ret})\n        end\n      end\n\n      def run_choreography(impl, config) do\n        if function_exported?(impl, :run_choreography, 2) do\n          impl.run_choreography(impl, config)\n        else\n          send(config[Bob], impl.pick_modulus())\n\n          (\n            bob_key =\n              receive do\n                msg -> msg\n              end\n\n            impl.encrypt(message, bob_key)\n          )\n        end\n      end\n    end\n  )\n\n  (\n    def bob do\n      quote do\n        import Bob\n        @behaviour Bob\n        def init() do\n          Bob.init(__MODULE__)\n        end\n      end\n    end\n\n    defmodule Bob do\n      @callback gen_key(any()) :: any()\n      def init(impl) do\n        receive do\n          {:config, config} ->\n            ret = run_choreography(impl, config)\n            send(config[:super], {:chorex_return, Bob, ret})\n        end\n      end\n\n      def run_choreography(impl, config) do\n        if function_exported?(impl, :run_choreography, 2) do\n          impl.run_choreography(impl, config)\n        else\n          m =\n            receive do\n              msg -> msg\n            end\n\n          send(config[Alice], impl.gen_key(m))\n        end\n      end\n    end\n  )\n\n  defmacro __using__(which) do\n    apply(__MODULE__, which, [])\n  end\nend\n```\n\nYou can see there's a `Chorex.Alice` module and a `Chorex.Bob` module.","ref":"readme.html#high-level-internals"},{"type":"extras","title":"Testing - README","doc":"Simply clone the repository and run `mix test`.\n\n\n  \n\n# Bibliography\n\n-   Hirsch & Garg (2022-01-16) *Pirouette: Higher-Order Typed Functional Choreographies*, Proceedings of the ACM on Programming Languages.  \n\n-   Lugović & Montesi (2023-10-15) *Real-World Choreographic Programming: Full-Duplex Asynchrony and Interoperability*, The Art, Science, and Engineering of Programming.  \n\n\n# Authors\n\nThis is a project by the [Utah PLT](https://github.com/utahplt) group. Primary development by [Ashton Wiersdorf](https://lambdaland.org).","ref":"readme.html#testing"}],"content_type":"text/markdown","producer":{"name":"ex_doc","version":[48,46,51,52,46,48]}}